"""
AyxPlugin (required) has-a IncomingInterface (optional).
Although defining IncomingInterface is optional, the interface methods are needed if an upstream tool exists.
"""

import AlteryxPythonSDK as Sdk
import xml.etree.ElementTree as Et
import dateparser

class AyxPlugin:
    """
    Implements the plugin interface methods, to be utilized by the Alteryx engine to communicate with a plugin.
    Prefixed with "pi", the Alteryx engine will expect the below five interface methods to be defined.
    """

    def __init__(self, n_tool_id: int, alteryx_engine: object, output_anchor_mgr: object):
        """
        Constructor is called whenever the Alteryx engine wants to instantiate an instance of this plugin.
        :param n_tool_id: The assigned unique identification for a tool instance.
        :param alteryx_engine: Provides an interface into the Alteryx engine.
        :param output_anchor_mgr: A helper that wraps the outgoing connections for a plugin.
        """

        # Default properties
        self.n_tool_id = n_tool_id
        self.alteryx_engine = alteryx_engine
        self.output_anchor_mgr = output_anchor_mgr
        self.is_initialized = False

        # Custom properties
        self.datefield = None
        self.datetype = None
        self.output_anchor = None

    def pi_init(self, str_xml: str):
        """
        Handles building out the sort info, to pass into pre_sort() later on, from the user configuration.
        Called when the Alteryx engine is ready to provide the tool configuration from the GUI.
        :param str_xml: The raw XML from the GUI.
        """

        # Initialize

        if self.alteryx_engine.get_init_var(self.n_tool_id, 'UpdateOnly') == 'True':
            self.is_initialized = False
            return None
        else:
            self.is_initialized = True


        # Getting the user-entered selections from the GUI
        if Et.fromstring(str_xml).find('FieldSelect') is not None:
            self.datafield = Et.fromstring(str_xml).find('FieldSelect').text
            if self.datafield == None:
                self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.error, 'Please select field to parse')
                self.is_initialized = False

        if Et.fromstring(str_xml).find('TypeSelect') is not None:
            self.output_type = Et.fromstring(str_xml).find('TypeSelect').text

            if self.output_type == "Date":
                self.datetype = Sdk.FieldType.date
            else:
                self.datetype = Sdk.FieldType.datetime
        else:
            self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.error, 'Please select an output type')

        self.output_anchor = self.output_anchor_mgr.get_output_anchor('Output')  # Getting the output anchor from the XML file.

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> object:
        """
        The IncomingInterface objects are instantiated here, one object per incoming connection, also pre_sort() is called here.
        Called when the Alteryx engine is attempting to add an incoming data connection.
        :param str_type: The name of the input connection anchor, defined in the Config.xml file.
        :param str_name: The name of the wire, defined by the workflow author.
        :return: The IncomingInterface object(s).
        """

        self.single_input = IncomingInterface(self)
        return self.single_input

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        """
        Called when the Alteryx engine is attempting to add an outgoing data connection.
        :param str_name: The name of the output connection anchor, defined in the Config.xml file.
        :return: True signifies that the connection is accepted.
        """

        return True

    def pi_push_all_records(self, n_record_limit: int) -> bool:
        """
        Called when a tool has no incoming data connection.
        :param n_record_limit: Set it to <0 for no limit, 0 for no records, and >0 to specify the number of records.
        :return: True for success, False for failure.
        """

        self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.error, 'Missing Incoming Connection')
        return False

    def pi_close(self, b_has_errors: bool):
        """
        Called after all records have been processed..
        :param b_has_errors: Set to true to not do the final processing.
        """
        if self.output_anchor != None:
            self.output_anchor.assert_close()  # Checks whether connections were properly closed.



class IncomingInterface:
    """
    This optional class is returned by pi_add_incoming_connection, and it implements the incoming interface methods, to
    be utilized by the Alteryx engine to communicate with a plugin when processing an incoming connection.
    Prefixed with "ii", the Alteryx engine will expect the below four interface methods to be defined.
    """

    def __init__(self, parent: object):
        """
        Constructor for IncomingInterface.
        :param parent: AyxPlugin
        """

        # Default properties
        self.parent = parent

        # Custom properties


    def ii_init(self, record_info_in: object) -> bool:
        """
        Called to report changes of the incoming connection's record metadata to the Alteryx engine.
        :param record_info_in: A RecordInfo object for the incoming connection's fields.
        :return: True for success, otherwise False.
        """
        # Checks if initialized
        if not self.parent.is_initialized:
            return False



        record_info_out = record_info_in.clone()  # Since no new data is being introduced, setting the outgoing layout the same as record_info_in.

        # Create output field

        record_info_out.add_field(self.parent.datafield + "_BBDate", self.parent.datetype)
        self.parent.bbdate_field = record_info_out[record_info_out.get_field_num(self.parent.datafield + "_BBDate")]

        # Create record creator and copier
        self.record_creator = record_info_out.construct_record_creator()
        self.record_copier = Sdk.RecordCopier(record_info_out, record_info_in)



        # Something important

        for index in range(record_info_in.num_fields):
            self.record_copier.add(index, index)

        self.record_copier.done_adding()

        # Get hook to the input field
        self.parent.input_field = record_info_out[record_info_out.get_field_num(self.parent.datafield)]

        self.parent.output_anchor.init(record_info_out)  # Lets the downstream tools know what the outgoing record metadata will look like, based on record_info_out.
        return True

    def ii_push_record(self, in_record: object) -> bool:
        """
        Responsible for pushing records out, under a count limit set by the user in n_record_select.
        Called when an input record is being sent to the plugin.
        :param in_record: The data for the incoming record.
        :return: False if method calling limit (record_cnt) is hit.
        """

        #Copy the existing data over
        self.record_creator.reset()
        self.record_copier.copy(self.record_creator, in_record)

        #Do the conversion
        self.date = self.parent.input_field.get_as_string(in_record)
        if self.date is not None:
            self.parsed = dateparser.parse(self.date)
        else:
            self.parsed = None
        if self.parsed is not None:
            self.parsed = self.parsed.isoformat()[:19].replace("T", " ")
            self.parent.bbdate_field.set_from_string(self.record_creator, self.parsed)
        else:
            self.parent.bbdate_field.set_null(self.record_creator)

        # Finalize record
        out_record = self.record_creator.finalize_record()


        # Push
        if not self.parent.output_anchor.push_record(out_record):
            return False

        return True

    def ii_update_progress(self, d_percent: float):
        """
        Called by the upstream tool to report what percentage of records have been pushed.
        :param d_percent: Value between 0.0 and 1.0.
        """

        self.parent.alteryx_engine.output_tool_progress(self.parent.n_tool_id, d_percent)  # Inform the Alteryx engine of the tool's progress.
        self.parent.output_anchor.update_progress(d_percent)  # Inform the downstream tool of this tool's progress.

    def ii_close(self):
        """
        Called when the incoming connection has finished passing all of its records.
        """

        self.parent.output_anchor.output_record_count(True)  # True: Let Alteryx engine know that all records have been sent downstream.
        self.parent.output_anchor.close()  # Close outgoing connections.
